local LrXml = import "LrXml"
local LrHttp = import "LrHttp"
local LrDialogs = import "LrDialogs"
local LrView = import "LrView"
local LrDate = import "LrDate"
local LrColor = import "LrColor"
local LrTasks = import "LrTasks"
local LrLogger = import "LrLogger"
local LrPathUtils = import "LrPathUtils"
local LrApplication = import "LrApplication"
local LrPrefs = import "LrPrefs"
local LrStringUtils = import "LrStringUtils"
local LrMD5 = import "LrMD5"

UpdateCheck = {}

local CheckVersion = nil

UpdateCheck.vInfo = require("Info.lua")
local log = LrLogger(UpdateCheck.vInfo.logname)

UpdateCheck.SafeParseXml = function( textXml )
	local Dom = nil
	if not pcall(function() Dom = LrXml.parseXml(textXml) end) then
		if not pcall(function() Dom = LrXml.parseXml(textXml:gsub("&#x[a-fA-F0-9]-;","")) end) then
			log:info("Unreadable XML contents follow:")
			log:info(textXml)
			LrDialogs.message("Plugin Error","Unable to read XML contents.","critical")
		end
	end
	
	return Dom
end

function UpdateCheck.DumpHeaders ( Headers, CustomMsg )
	log:trace("Server status: " .. (Headers["status"] or "(nil)"))
	log:trace("Lightroom: " .. LrApplication.versionString() );
	log:trace("Headers:")
	
	local ErrMsg = "Unknown error"
	
	if Headers["status"] then
		ErrMsg = "The server said '" .. Headers["status"] .. "'."
	end
	
	for field, hvalue in pairs(Headers) do
		if type(hvalue) ~= "table" then
			log:trace((field or "(nil)") .. " = " .. (hvalue or "(nil)") )
		end
	end

	log:trace("Errors:")
	local errtab = Headers["error"]
	if errtab then
		for field, value in pairs(errtab) do
			log:trace("errtab " .. (field or "(nil)") .. " = " .. (value or "(nil)") )
			if field == "name"  then
				ErrMsg = value
			end
		end
	else
		log:trace("No error table")
	end
	
	if CustomMsg then
		LrDialogs.message(CustomMsg,ErrMsg,"critical")
	end
end

UpdateCheck.UpdateCheck = function(ShowIfNone, prefs)
	-- Check for updates
	log:enable("logfile")

	-- Handle being called from a webengine where LrPrefs is not relevant.
	if _PLUGIN ~= nil and prefs == nil then
		prefs = LrPrefs.prefsForPlugin()
	end
	
	local response = nil
	local UpVers = nil
	local InfoVersion = UpdateCheck.vInfo.VERSION
	local InfoName = UpdateCheck.vInfo.LrPluginName
	local UpdateName = UpdateCheck.vInfo.webUpdateName
	local ThisVersion = InfoVersion.major .. "." .. InfoVersion.minor .. "." .. InfoVersion.revision

	response, Headers = LrHttp.get( "http://updates.jimkeir.co.uk/" .. UpdateCheck.vInfo.webUpdateName .."-"..InfoVersion.major..".xml", nil, 30 )
	if response then
		if Headers["status"] == 200 then
			local Dom = UpdateCheck.SafeParseXml(response)
			if Dom then
				if Dom:name() == "Update" then
					-- Looks more or less OK.
					local UpMsg = nil
					local UpUrl = nil
					local UpDate = nil
					
					-- Populate the expiry value
					UpdateCheck.StringLookup(prefs) 

					local Atts = Dom:attributes()
					if Atts ~= nil then
						UpUrl = Atts["url"].value
						UpVers = Atts["version"].value
						UpDate = (Atts["reldate"] and LrDate.timeFromPosixDate(Atts["reldate"].value) or LrDate.currentTime())
					end
					if Dom:childCount() > 0 then
						UpMsg = Dom:childAtIndex(1):text()
						UpMsg = UpMsg:gsub("\n","\r\n")
					end

					if UpUrl ~= nil and UpVers ~= nil and UpMsg ~= nil then
						local uvParse = {}
						for w in (UpVers .. "."):gmatch("([^.]*).") do 
							table.insert(uvParse, w)
							if #uvParse == 3 then break end
						end
						UpVers = table.concat(uvParse, ".")
						
						if ThisVersion < UpVers then
							local f = LrView.osFactory()
							local c = f:column {
								spacing = f:control_spacing(),
								
								f:row {
									f:static_text {
--										text_color = LrColor(1,0,0),
										title = "You are using version " .. ThisVersion,
									},
								},
								f:row {
									f:static_text {
										text_color = LrColor(1,0,0),
										title = "An update to version " .. UpVers .. " is available!",
									},
								},
								
								f:edit_field {
									value = UpMsg,
									height_in_lines = 4,
									width_in_chars = 40,
									enabled = false,
								},

								f:row {
									f:static_text {
										title = "URL:",
									},
									f:edit_field {
										value = UpUrl,
										height_in_lines = 1,
										width_in_chars = 36,
										enabled = false,
									},
								},
								
								f:row {
									spacing = f:control_spacing(),
									f:push_button {
										width = LrView:share("dlbtn"),
										title = "Download",
										enabled = UpDate < (prefs.Expiry or 0),
										action = function()
													LrHttp.openUrlInBrowser(UpUrl)
												 end
									},

									f:static_text {
										title = "Your license allows you free updates until " .. LrDate.formatLongDate(prefs.Expiry) .. "\n" ..
												"The update was released on " .. LrDate.formatLongDate(UpDate),
										width_in_chars = 32,
										height_in_lines = 2,
										visible = UpDate >= (prefs.Expiry or 0),
									},
								},
							}
							LrDialogs.presentModalDialog( { title = InfoName, contents = c, } )
						elseif ShowIfNone then
							LrDialogs.message("No update available","You are running the latest version.","info")
						end
					else
						log:trace("Update XML doesn't contain enough information.")
						LrDialogs.message("Update XML doesn't contain enough information.","HTTP error", "critical")
					end
				end
			else
				log:trace("Update XML invalid.")
				LrDialogs.message("Update information is not valid.","HTTP error ", "critical")
			end
		elseif Headers["status"] == 404 then
			LrDialogs.message("Update","No update information was available. Please try again later.")
		else
			log:trace("Failed update fetch : " .. Headers["status"] )
			UpdateCheck.DumpHeaders(Headers, "Unable to fetch update information.")
		end
	else
		UpdateCheck.DumpHeaders(Headers, "Unable to fetch update information.")
	end
	
	return UpVers or "unknown"
end

UpdateCheck.URLEncode = function(str)
  if (str) then
    str = string.gsub (str, "\n", "\r\n")
    str = string.gsub (str, "([^%w %-%_%.%~])",
        function (c) return string.format ("%%%02X", string.byte(c)) end)
    str = string.gsub (str, " ", "+")
  end
  
  return str	
end

-- This propertyTable is *not* the same as the plugin prefs. Need to explicitly load prefs.
local UpdateNowDialog = function(viewFactory, propertyTable)
	local prefs = nil

	if _PLUGIN ~= nil and prefs == nil then
		prefs = LrPrefs.prefsForPlugin()
	else
		-- Only for non-plugins where this function is NOT called directly from Lightroom itself.
		prefs = propertyTable
	end

	-- Populate the expiry value
	UpdateCheck.StringLookup(prefs) 

	ExpiryValue = 
		viewFactory:static_text {
			title = LrView.bind {
				keys = { "Expiry" },
				bind_to_object = prefs,
				operation = function(binder, values, fromTable)
					if fromTable then
						if values.Expiry > 0 then
							if LrDate.timeFromPosixDate((UpdateCheck.vInfo.expval or 0)) < values.Expiry then
								ExpiryValue.text_color = LrColor(0,0,0)
							else
								ExpiryValue.text_color = LrColor(1,0,0)
							end

							return LrDate.formatLongDate(values.Expiry)
						else
							return "Trial Copy"
						end
					else
						return LrBinding.kUnsupportedDirection
					end
				end,
			},
			width_in_chars = 16,
		}

	CheckVersion = viewFactory:static_text { title = "", width_in_chars = 32, }
		
	return {
		{
			title = "License Details",
			bind_to_object = prefs,
			
			viewFactory:column {
				spacing = viewFactory:control_spacing(),
				
				viewFactory:row {
					spacing = viewFactory:control_spacing(),
					viewFactory:push_button {
						title = "Check for Updates",
						action = function(button)
							LrTasks.startAsyncTask( function()
									button.enabled = false	
									CheckVersion.title = "Checking..."
									CheckVersion.title = "Latest is " .. UpdateCheck.UpdateCheck(true, prefs)
									button.enabled = true
								end )
							end,
					},
					CheckVersion,
				},

				viewFactory:row {
					spacing = viewFactory:control_spacing(),
					viewFactory:static_text {
						title = "e-Mail",
						width_in_chars = 10,
					},
					viewFactory:edit_field {
						value = LrView.bind {
							keys = { "UserName" },
							transform = function(value, fromTable)
								if not fromTable then
									if prefs.UserName ~= value then
										prefs.UserName = value
										
										UpdateNowFinish(prefs)
										if UpdateCheck.StringLookup(prefs) > 0 then
											LrDialogs.message("License Information", "Thank you for licensing this plugin!")
										end
									end
								else
									value = (prefs.UserName or ""):match "^%s*(.-)%s*$"
								end

								return value
							end,
						},
						
						immediate = false,
						width_in_chars = 32,
					},
				},

				viewFactory:row {
					spacing = viewFactory:control_spacing(),
					
					viewFactory:static_text {
						title = "Plug-in License",
						width_in_chars = 10,
					},
					viewFactory:edit_field {
						value = LrView.bind {
							keys = { "License" },
							transform = function(value, fromTable)
								if not fromTable then
									if prefs.License ~= value then
										prefs.License = value
										
										UpdateNowFinish(prefs)
										if UpdateCheck.StringLookup(prefs) > 0 then
											LrDialogs.message("License Information", "Thank you for licensing this plugin!")
										end
									end
								else
									value = (prefs.License or ""):match "^%s*(.-)%s*$"
								end

								return value
							end,
						},

						immediate = false,
						width_in_chars = 32,
					},
				},

				viewFactory:row {
					viewFactory:column {

						spacing = viewFactory:control_spacing(),
						viewFactory:row {
							spacing = viewFactory:control_spacing(),

							viewFactory:static_text {
								title = "This Version",
								width_in_chars = 10,
							},

							viewFactory:static_text {
								title = "V" .. UpdateCheck.vInfo.VERSION.major .. "." .. UpdateCheck.vInfo.VERSION.minor .. "." .. UpdateCheck.vInfo.VERSION.revision .. 
											", released " .. LrDate.formatLongDate(LrDate.timeFromPosixDate((UpdateCheck.vInfo.expval or 0))),
								width_in_chars = 32,
							},						
						},
						viewFactory:row {
							spacing = viewFactory:control_spacing(),

							viewFactory:static_text {
								title = "Free updates until",
								width_in_chars = 10,
							},

							ExpiryValue,
						},
					},
					
					viewFactory:column {
						viewFactory:view {
							visible = LrView.bind {
								keys = { "Expiry" },
								bind_to_object = prefs,
								operation = function(binder, values, fromTable)
									if fromTable then
										return values.Expiry == 0
									else
										return LrBinding.kUnsupportedDirection
									end
								end,
							},

							viewFactory:static_text {
								title = "The license and email must *both* be entered\nexactly as you received them, including\nsmall and capital letters.",
								width_in_chars = 30,
								height_in_lines = 3,
								text_color = LrColor(1,0,0),
							},
						},
					},
				},

				viewFactory:row {
					spacing = viewFactory:control_spacing(),
					viewFactory:push_button {
						title = "Retrieve License",
						enabled = LrView.bind {
							keys = { "UserName" },
							bind_to_object = prefs,
							operation = function( binder, values, fromTable )
									if fromTable then
										return (values.UserName or ''):find('@') ~= nil
									else
										return LrBinding.kUnsupportedDirection
									end
								end,
						},
						action = function(button)
							LrTasks.startAsyncTask( function()
									button.enabled = false	

									-- Do the retrieve
									local postBody = "email=" .. UpdateCheck.URLEncode(prefs["UserName"]) ..
													 "&product=" .. UpdateCheck.URLEncode((UpdateCheck.vInfo.retrieveName or UpdateCheck.vInfo.LrPluginName))

									response, Headers = LrHttp.post( "https://www.lightroom-plugins.com/support/index.php?/Lightroom/vshop/Retrieve", 
																postBody, { { field = "Content-Type", value = "application/x-www-form-urlencoded" } }, 30 )
									if response then
										if Headers["status"] == 200 then
											LrDialogs.message("License Retrieval", response)
										else
											LrDialogs.message("License Retrieval", "There was a problem requesting the license. Please open a support ticket.")
										end
									else
										LrDialogs.message("License Retrieval", "Unable to contact the support server. Please try again later.")
									end

									button.enabled = true	
								end )
							end,
					},
					
					viewFactory:push_button {
						title = LrView.bind {
							keys = { "UserName", "License" },
							bind_to_object = prefs,
							operation = function( binder, values, fromTable )
									if fromTable then
							 			local vCheck = UpdateCheck.StringLookup(prefs)
										if vCheck == 0 then
											return "Buy License"
										end
										return "Extend License"
									else
										return LrBinding.kUnsupportedDirection
									end
								end,
						},
						enabled = true,
						action = function(button)
                                if (prefs.userName or '') ~= '' then
                                    LrDialogs.message("Get New License","You will need to log into the Support Center using the email address " ..
                                        "'" .. prefs.UserName .. "' and the correct password.\n\nIf you don't know your password, please use the 'Lost Password' " ..
                                        "link in the Support Center to have it emailed to you.")
                                end

								LrHttp.openUrlInBrowser("https://www.lightroom-plugins.com/support/index.php?/Lightroom/vshop/Buy/Main/" ..
										UpdateCheck.vInfo.webUpdateName)
							end,
					},
				},
			},
		},
		
		type(UpdateCheck.vInfo.settingsSections) == "function" and UpdateCheck.vInfo.settingsSections(LrView, prefs) or nil,
	}
end

function RotateN(n, t)
  local byte_a, byte_A = string.byte('a'), string.byte('A')

  return (string.gsub(t, ".",
      function (char)
        local offset = (char >= 'a') and byte_a or byte_A
        local b = string.byte(char) - offset -- 0 to 25
        if b >= 0 and b < 26 then
			b = math.mod(b + n, 26) + offset
		else
			b = string.byte(char)
		end
		
		return string.char(b)
      end
    ))
end

function string.fromhex(str)
    return (str:gsub('..', function (cc)
        return string.char(tonumber(cc, 16))
    end))
end

function string.tohex(str)
    return (str:gsub('.', function (c)
        return string.format('%02X', string.byte(c))
    end))
end

-- return single bit (for OR)
function bit(x,b)
	return (math.mod(x, 2^b) - math.mod(x, 2^(b-1)) > 0)
end

-- logic XOR for number values
function xor(x,y)
	local result = 0

	for p = 1, 8 do 
		local b1 = bit(x, p)
		local b2 = bit(y, p)
		
		result = result + (((b1 or b2) and (b1 ~= b2)) and 2^(p-1) or 0) 
	end

	return result
end

function UpdateCheck.StringLookup(propertyTable)
	-- Take the user's account name and one of a set of special strings and encode them.
	-- Fairly simple, but never mind...
	
	local isValid = false
	
	local ServiceLevel = {
		RotateN(math.mod(UpdateCheck.vInfo.VERSION.revision, 26), UpdateCheck.vInfo.retval),
	}

	--[[
		- Format 1 : License == LrMD5(email .. secret)
		- Format 2 : License == LrStringUtils.encodeBase64(LrMD5(expiry .. email .. secret):fromhex() .. expiry)
	]]--
	
	-- Default to 01 March 2015
	propertyTable.Expiry = 0 -- 1425168000 - 978307200

	local TryThis = 1
	while ServiceLevel[TryThis] do
		-- V1 license check
		local WorkString = LrStringUtils.trimWhitespace(propertyTable.UserName or "") .. LrStringUtils.decodeBase64(ServiceLevel[TryThis])
		WorkString = LrStringUtils.upper(LrMD5.digest(WorkString))
		
--		log:trace('Compare "' .. WorkString .. '" with "' .. LrStringUtils.trimWhitespace(propertyTable.License) .. '"')
		if WorkString == LrStringUtils.trimWhitespace(propertyTable.License or "") then
--			log:trace('Allowing ' .. LrStringUtils.decodeBase64(ServiceLevel[TryThis]) .. '( == ' .. TryThis .. ')')

			-- Default to 01 March 2015
			propertyTable.Expiry = 1425168000 - 978307200

			-- TODO: Fetch an updated license from the server.
			isValid = true
			break
		end
		
		-- V2 license check
		local bytes = LrStringUtils.decodeBase64(LrStringUtils.trimWhitespace(propertyTable.License or ""))
		
		if bytes ~= nil and bytes:len() == 24 then
			-- First 16 bytes are binary md5, last 8 are hex expiry date
			local md5 = bytes:sub(1, 16)
			local ts  = bytes:sub(17)
			local tsText = ''
			
			for i = 1, 8 do
				local c = xor(ts:byte(i), md5:byte(1 + ((i-1) * 2)))
				tsText = tsText .. string.char(c)
			end

--			LrDialogs.message("Timestamp = " .. tsText)
			
			ts = tonumber(tsText, 16)

			if ts ~= nil then
--				LrDialogs.message("Decimal Timestamp = " .. ts)
				md5 = string.tohex(md5)

				WorkString = LrStringUtils.trimWhitespace(propertyTable.UserName or "") .. LrStringUtils.decodeBase64(ServiceLevel[TryThis]) .. ts
				if md5 == LrStringUtils.upper(LrMD5.digest(WorkString)) then
					local dt = LrDate.timeFromPosixDate(ts)
					
					propertyTable.Expiry = dt
--					LrDialogs.message("V2 license accepted", "Expires on " .. ts .. " (" .. LrDate.formatLongDate(dt) .. ")")
					isValid = true
					break
				end
			end
		end
		
		TryThis = TryThis + 1
	end

	if isValid then
--	LrDialogs.message("Release: " .. LrDate.formatLongDate(LrDate.timeFromPosixDate(UpdateCheck.vInfo.expval or 0)) .. "; " ..
--					  "Expiry: " .. LrDate.formatLongDate(propertyTable.Expiry))

		if LrDate.timeFromPosixDate((UpdateCheck.vInfo.expval or 0)) < propertyTable.Expiry then
			return TryThis
		else
			return -1
		end
	end

	return 0
end

-- propertyTable here is *not* prefs...
function UpdateNowFinish(propertyTable)
	if _PLUGIN == nil or _PLUGIN.enabled then

		-- Handle being called from a webengine where LrPrefs is not relevant.
		if _PLUGIN ~= nil then
			prefs = LrPrefs.prefsForPlugin()
		else
			prefs = propertyTable
		end

		if propertyTable["License"] ~= nil and propertyTable["License"] ~= '' then
			local vCheck = UpdateCheck.StringLookup(propertyTable)
			if vCheck == 0 then
				LrDialogs.message(UpdateCheck.vInfo.retrieveName, "Your license is not valid. "..
									"Please check it in the email you received with the license details.")
			elseif vCheck == -1 then
				LrDialogs.message(UpdateCheck.vInfo.retrieveName, "Your license is valid " ..
									"but expired before this version of the software was released.\n\n" ..
									"Please restore a backup of the old software, or purchase a support extension.")
			end
		end
	end
end

return {
	endDialog = UpdateNowFinish,
	sectionsForTopOfDialog = UpdateNowDialog, 
}
